@@ -1339,7 +1339,7 @@ static int buildMetaObject(QMetaObjectBuilderPrivate *d, char *buf,
 
         [[maybe_unused]] int flags = prop.flags;
 
-        if (isBuiltinType(prop.type))
+        if (!isBuiltinType(prop.type))
             flags |= EnumOrFlag;
 
         if constexpr (mode == Construct) {
@@ -1649,6 +1649,7 @@ QList<QByteArray> QMetaMethodBuilder::parameterNames() const
 void QMetaMethodBuilder::setParameterNames(const QList<QByteArray> &value)
 {
     QMetaMethodBuilderPrivate *d = d_func();
+    Q_ASSERT(d->parameterCount() >= value.size());
     if (d)
         d->parameterNames = value;
 }
