@@ -409,7 +409,7 @@ QQuaternion QQuaternion::fromAxisAndAngle
         (float x, float y, float z, float angle)
 {
     float length = qHypot(x, y, z);
-    if (!qFuzzyIsNull(length) && !qFuzzyCompare(length, 1.0f)) {
+    if (!qFuzzyCompare(length, 1.0f) && !qFuzzyIsNull(length)) {
         x /= length;
         y /= length;
         z /= length;
@@ -448,19 +448,71 @@ QQuaternion QQuaternion::fromAxisAndAngle
 #endif // QT_NO_VECTOR3D
 
 /*!
+    \fn void QQuaternion::getEulerAngles(float *pitch, float *yaw, float *roll) const
     \since 5.5
 
+    \obsolete
+
+    Use eulerAngles() instead.
+
     Calculates \a roll, \a pitch, and \a yaw Euler angles (in degrees)
     that corresponds to this quaternion.
 
     All of \a pitch, \a yaw, and \a roll must be valid, non-\nullptr pointers,
     otherwise the behavior is undefined.
 
+    \sa eulerAngles(), fromEulerAngles()
+*/
+
+/*!
+    \since 6.11
+    \class QQuaternion::EulerAngles
+    \ingroup painting-3D
+    \inmodule QtGui
+
+    A struct containing three fields \l{pitch}, \l{yaw}, and \l{roll},
+    representing the three Euler angles that define a
+    \l{QQuaternion}{quaternion}.
+
+    Consult the documentation of functions taking or returning an EulerAngles
+    object for the order in which the rotations are applied.
+
+    \sa QQuaternion::eulerAngles(), QQuaternion::fromEulerAngles(QQuaternion::EulerAngles<float>)
+*/
+
+/*!
+    \variable QQuaternion::EulerAngles::pitch
+
+    The pitch represents the rotation around the x-axis.
+*/
+
+/*!
+    \variable QQuaternion::EulerAngles::yaw
+
+    The yaw represents the rotation around the y-axis.
+*/
+
+/*!
+    \variable QQuaternion::EulerAngles::roll
+
+    The roll represents the rotation around the z-axis.
+*/
+
+/*!
+    \since 6.11
+
+    Returns the Euler angles (in degrees) that correspond to this quaternion.
+
     \sa fromEulerAngles()
 */
-void QQuaternion::getEulerAngles(float *pitch, float *yaw, float *roll) const
+auto QQuaternion::eulerAngles() const -> EulerAngles<float>
 {
-    Q_ASSERT(pitch && yaw && roll);
+    EulerAngles<float> result;
+
+    // to avoid churn
+    auto pitch = &result.pitch;
+    auto yaw = &result.yaw;
+    auto roll = &result.roll;
 
     // Algorithm adapted from:
     // https://ingmec.ual.es/~jlblanco/papers/jlblanco2010geometry3D_techrep.pdf
@@ -508,6 +560,8 @@ void QQuaternion::getEulerAngles(float *pitch, float *yaw, float *roll) const
     *pitch = qRadiansToDegrees(*pitch);
     *yaw = qRadiansToDegrees(*yaw);
     *roll = qRadiansToDegrees(*roll);
+
+    return result;
 }
 
 /*!
@@ -517,7 +571,7 @@ void QQuaternion::getEulerAngles(float *pitch, float *yaw, float *roll) const
     \a roll degrees around the z axis, \a pitch degrees around the x axis,
     and \a yaw degrees around the y axis (in that order).
 
-    \sa getEulerAngles()
+    \sa eulerAngles(), toEulerAngles(), fromEulerAngles(QQuaternion::EulerAngles<float>)
 */
 QQuaternion QQuaternion::fromEulerAngles(float pitch, float yaw, float roll)
 {
@@ -549,6 +603,19 @@ QQuaternion QQuaternion::fromEulerAngles(float pitch, float yaw, float roll)
     return QQuaternion(w, x, y, z);
 }
 
+/*!
+    \fn QQuaternion QQuaternion::fromEulerAngles(EulerAngles<float> angles)
+    \since 6.11
+    \overload
+
+    Equivalent to
+    \code
+    fromEulerAngles(angles.pitch, angles.yaw, angles.roll);
+    \endcode
+
+    \sa eulerAngles(), toEulerAngles(), fromEulerAngles()
+*/
+
 /*!
     \since 5.5
 
@@ -557,7 +624,7 @@ QQuaternion QQuaternion::fromEulerAngles(float pitch, float yaw, float roll)
     \note If this quaternion is not normalized,
     the resulting rotation matrix will contain scaling information.
 
-    \sa fromRotationMatrix(), getAxes()
+    \sa fromRotationMatrix(), toAxes()
 */
 QMatrix3x3 QQuaternion::toRotationMatrix() const
 {
@@ -640,8 +707,63 @@ QQuaternion QQuaternion::fromRotationMatrix(const QMatrix3x3 &rot3x3)
 #ifndef QT_NO_VECTOR3D
 
 /*!
+    \since 6.11
+    \class QQuaternion::Axes
+    \ingroup painting-3D
+    \inmodule QtGui
+
+    A struct containing the three orthonormal axes that define a
+    \l{QQuaternion}{quaternion}.
+
+
+    \sa QQuaternion::toAxes(), QQuaternion::fromAxes(QQuaternion::Axes)
+*/
+
+/*!
+    \variable QQuaternion::Axes::x
+
+    The x orthonormal axis that, together with \l{y} and \l{z}, defines a
+    quaternion.
+*/
+
+/*!
+    \variable QQuaternion::Axes::y
+
+    The y orthonormal axis that, together with \l{x} and \l{z}, defines a
+    quaternion.
+*/
+
+/*!
+    \variable QQuaternion::Axes::z
+
+    The z orthonormal axis that, together with \l{x} and \l{y}, defines a
+    quaternion.
+*/
+
+/*!
+    \since 6.11
+
+    Returns the three orthonormal axes that define this quaternion.
+
+    \sa QQuaternion::Axes, fromAxes(QQuaternion::Axes), toRotationMatrix()
+*/
+auto QQuaternion::toAxes() const -> Axes
+{
+    const QMatrix3x3 rot3x3(toRotationMatrix());
+
+    return { {rot3x3(0, 0), rot3x3(1, 0), rot3x3(2, 0)},
+             {rot3x3(0, 1), rot3x3(1, 1), rot3x3(2, 1)},
+             {rot3x3(0, 2), rot3x3(1, 2), rot3x3(2, 2)} };
+}
+
+
+/*!
+    \fn void QQuaternion::getAxes(QVector3D *xAxis, QVector3D *yAxis, QVector3D *zAxis) const
     \since 5.5
 
+    \obsolete
+    Use toAxes() instead.
+
     Returns the 3 orthonormal axes (\a xAxis, \a yAxis, \a zAxis) defining the quaternion.
 
     All of \a xAxis, \a yAxis, and \a zAxis must be valid, non-\nullptr pointers,
@@ -649,16 +771,6 @@ QQuaternion QQuaternion::fromRotationMatrix(const QMatrix3x3 &rot3x3)
 
     \sa fromAxes(), toRotationMatrix()
 */
-void QQuaternion::getAxes(QVector3D *xAxis, QVector3D *yAxis, QVector3D *zAxis) const
-{
-    Q_ASSERT(xAxis && yAxis && zAxis);
-
-    const QMatrix3x3 rot3x3(toRotationMatrix());
-
-    *xAxis = QVector3D(rot3x3(0, 0), rot3x3(1, 0), rot3x3(2, 0));
-    *yAxis = QVector3D(rot3x3(0, 1), rot3x3(1, 1), rot3x3(2, 1));
-    *zAxis = QVector3D(rot3x3(0, 2), rot3x3(1, 2), rot3x3(2, 2));
-}
 
 /*!
     \since 5.5
@@ -667,7 +779,7 @@ void QQuaternion::getAxes(QVector3D *xAxis, QVector3D *yAxis, QVector3D *zAxis)
 
     \note The axes are assumed to be orthonormal.
 
-    \sa getAxes(), fromRotationMatrix()
+    \sa toAxes(), fromRotationMatrix()
 */
 QQuaternion QQuaternion::fromAxes(const QVector3D &xAxis, const QVector3D &yAxis, const QVector3D &zAxis)
 {
@@ -685,6 +797,17 @@ QQuaternion QQuaternion::fromAxes(const QVector3D &xAxis, const QVector3D &yAxis
     return QQuaternion::fromRotationMatrix(rot3x3);
 }
 
+/*!
+    \since 6.11
+    \overload
+
+    \sa toAxes(), fromRotationMatrix()
+*/
+QQuaternion QQuaternion::fromAxes(Axes axes) // clazy:exclude=function-args-by-ref
+{
+    return fromAxes(axes.x, axes.y, axes.z);
+}
+
 /*!
     \since 5.5
 
