@@ -18,6 +18,7 @@
 #include "qwaylandplatformservices_p.h"
 #include "qwaylandscreen_p.h"
 #include "qwaylandcursor_p.h"
+#include "qwaylandeventdispatcher_p.h"
 
 #if defined(Q_OS_MACOS)
 #  include <QtGui/private/qcoretextfontdatabase_p.h>
@@ -25,7 +26,6 @@
 #else
 #  include <QtGui/private/qgenericunixfontdatabase_p.h>
 #endif
-#include <QtGui/private/qgenericunixeventdispatcher_p.h>
 #include <QtGui/private/qgenericunixtheme_p.h>
 
 #include <QtGui/private/qguiapplication_p.h>
@@ -88,6 +88,8 @@ QWaylandIntegration::QWaylandIntegration(const QString &platformName)
     : mPlatformName(platformName), mFontDb(new QGenericUnixFontDatabase())
 #endif
 {
+    QCoreApplication::setAttribute(Qt::AA_CompressHighFrequencyEvents);
+
     mDisplay.reset(new QWaylandDisplay(this));
     mPlatformServices.reset(new QWaylandPlatformServices(mDisplay.data()));
 
@@ -127,19 +129,20 @@ bool QWaylandIntegration::hasCapability(QPlatformIntegration::Capability cap) co
     case MultipleWindows:
     case NonFullScreenWindows:
         return true;
-    case RasterGLSurface:
-        return true;
     case WindowActivation:
         return true;
     case ScreenWindowGrabbing: // whether QScreen::grabWindow() is supported
         return false;
+    case OffscreenSurface:
+        return mDisplay->clientBufferIntegration()
+                && mDisplay->clientBufferIntegration()->canCreatePlatformOffscreenSurface();
     default: return QPlatformIntegration::hasCapability(cap);
     }
 }
 
 QPlatformWindow *QWaylandIntegration::createPlatformWindow(QWindow *window) const
 {
-    if ((window->surfaceType() == QWindow::OpenGLSurface || window->surfaceType() == QWindow::RasterGLSurface)
+    if (window->surfaceType() == QWindow::OpenGLSurface
         && mDisplay->clientBufferIntegration())
         return mDisplay->clientBufferIntegration()->createEglWindow(window);
 
@@ -159,6 +162,13 @@ QPlatformOpenGLContext *QWaylandIntegration::createPlatformOpenGLContext(QOpenGL
     return nullptr;
 }
 
+QPlatformOffscreenSurface *QWaylandIntegration::createPlatformOffscreenSurface(QOffscreenSurface *surface) const
+{
+    if (mDisplay->clientBufferIntegration())
+        return mDisplay->clientBufferIntegration()->createPlatformOffscreenSurface(surface);
+    return nullptr;
+}
+
 QOpenGLContext *QWaylandIntegration::createOpenGLContext(EGLContext context, EGLDisplay contextDisplay, QOpenGLContext *shareContext) const
 {
     return mClientBufferIntegration->createOpenGLContext(context, contextDisplay, shareContext);
@@ -172,7 +182,7 @@ QPlatformBackingStore *QWaylandIntegration::createPlatformBackingStore(QWindow *
 
 QAbstractEventDispatcher *QWaylandIntegration::createEventDispatcher() const
 {
-    return createUnixEventDispatcher();
+    return QWaylandEventDispatcher::createEventDispatcher();
 }
 
 QPlatformNativeInterface *QWaylandIntegration::createPlatformNativeInterface()
