@@ -64,9 +64,6 @@ private slots:
 void tst_QOpenGLWidget::initTestCase()
 {
     // See QOpenGLWidget constructor
-    if (!QGuiApplicationPrivate::platformIntegration()->hasCapability(QPlatformIntegration::RasterGLSurface))
-        QSKIP("QOpenGLWidget is not supported on this platform.");
-
     if (!QGuiApplicationPrivate::platformIntegration()->hasCapability(QPlatformIntegration::OpenGL))
         QSKIP("OpenGL is not supported on this platform.");
 }
